/*
 * Copyright(c) 2006 to 2021 ZettaScale Technology and others
 *
 * This program and the accompanying materials are made available under the
 * terms of the Eclipse Public License v. 2.0 which is available at
 * http://www.eclipse.org/legal/epl-2.0, or the Eclipse Distribution License
 * v. 1.0 which is available at
 * http://www.eclipse.org/org/documents/edl-v10.php.
 *
 * SPDX-License-Identifier: EPL-2.0 OR BSD-3-Clause
 */
#include <cstdlib>
#include <iostream>
#include <chrono>
#include <thread>

/* Include the C++ DDS API. */
#include "dds/dds.hpp"

/* Include data type and specific traits to be used with the C++ DDS API. */
#include "HelloWorldData.hpp"

using namespace org::eclipse::cyclonedds;


  class ExampleListener : public virtual dds::sub::DataReaderListener<HelloWorldData::Msg> {
  public:
      virtual void on_requested_deadline_missed (
          dds::sub::DataReader<HelloWorldData::Msg>& reader,
          const dds::core::status::RequestedDeadlineMissedStatus & status)
      {
          std::cout << "on_requested_deadline_missed" << std::endl;
      }
 
      virtual void on_requested_incompatible_qos (
          dds::sub::DataReader<HelloWorldData::Msg>& reader,
          const dds::core::status::RequestedIncompatibleQosStatus & status)
      {
          std::cout << "on_requested_incompatible_qos" << std::endl;
      }
 
      virtual void on_sample_rejected (
          dds::sub::DataReader<HelloWorldData::Msg>& reader,
          const dds::core::status::SampleRejectedStatus & status)
      {
          std::cout << "on_sample_rejected" << std::endl;
      }
 
      virtual void on_liveliness_changed (
          dds::sub::DataReader<HelloWorldData::Msg>& reader,
          const dds::core::status::LivelinessChangedStatus & status)
      {
          std::cout << "on_liveliness_changed" << std::endl;
      }
 
      virtual void on_data_available (
          dds::sub::DataReader<HelloWorldData::Msg>& reader)
      {
          std::cout << "on_data_available" << std::endl;
      }
 
      virtual void on_subscription_matched (
          dds::sub::DataReader<HelloWorldData::Msg>& reader,
          const dds::core::status::SubscriptionMatchedStatus & status)
      {
          std::cout << "on_subscription_matched, current_count: " << status.current_count() << std::endl;
      }
 
      virtual void on_sample_lost (
          dds::sub::DataReader<HelloWorldData::Msg>& reader,
          const dds::core::status::SampleLostStatus & status)
      {
          std::cout << "on_sample_lost" << std::endl;
      }
  };


int main() {

    try {
        std::cout << "=== [Subscriber] Create reader." << std::endl;

        dds::domain::DomainParticipant participant(domain::default_id());
        dds::topic::Topic<HelloWorldData::Msg> topic(participant, "HelloWorldData_Msg");
        dds::sub::Subscriber subscriber(participant);

        // Good case
        /*dds::sub::DataReader<HelloWorldData::Msg> reader(subscriber, topic,
            subscriber.default_datareader_qos(),
            new ExampleListener(),
            dds::core::status::StatusMask::all()
        );*/

        // Bad case
        dds::sub::DataReader<HelloWorldData::Msg> reader(subscriber, topic);
        auto *listener = new ExampleListener();
        reader.listener(listener, dds::core::status::StatusMask::all());


        std::cout << "=== [Subscriber] Wait for message." << std::endl;
        bool poll = true;
        while (poll) {
            dds::sub::LoanedSamples<HelloWorldData::Msg> samples;

            samples = reader.take();

            if (samples.length() > 0) {
                dds::sub::LoanedSamples<HelloWorldData::Msg>::const_iterator sample_iter;
                for (sample_iter = samples.begin();
                     sample_iter < samples.end();
                     ++sample_iter) {

                    const HelloWorldData::Msg& msg = sample_iter->data();
                    const dds::sub::SampleInfo& info = sample_iter->info();

                    if (info.valid()) {
                        std::cout << "=== [Subscriber] Message received:" << std::endl;
                        std::cout << "    userID  : " << msg.userID() << std::endl;
                        std::cout << "    Message : \"" << msg.message() << "\"" << std::endl;

                        poll = false;
                    }
                }
            } else {
                std::this_thread::sleep_for(std::chrono::milliseconds(20));
            }
        }

		// Just do make sure that we dont exit to fast and may miss a callback because of that?
        int ccc =0;
        while(ccc < 6) {
            std::this_thread::sleep_for(std::chrono::milliseconds(1000));
            ccc++;
        }

        //lisThread.join();
    } catch (const dds::core::Exception& e) {
        std::cerr << "=== [Subscriber] DDS exception: " << e.what() << std::endl;
        return EXIT_FAILURE;
    } catch (const std::exception& e) {
        std::cerr << "=== [Subscriber] C++ exception: " << e.what() << std::endl;
        return EXIT_FAILURE;
    }

    std::cout << "=== [Subscriber] Done." << std::endl;
    

    return EXIT_SUCCESS;
}